#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>

#if __GNUC__ >= 3
#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>
#endif

#include "column.h"
#include "head.h"
#include "util.h"

FitsColumn::FitsColumn(FitsHead* head, int i, int off)
{
  index_ = i;
  width_ = 0;
  offset_ = off;
  type_ = ' ';

  tform_ = head->getString(cat("TFORM",i));
  ttype_ = head->getString(cat("TTYPE",i));
  tunit_ = head->getString(cat("TUNIT",i));
  tscal_ = head->getReal(cat("TSCAL",i), 1);
  tzero_ = head->getReal(cat("TZERO",i), 0);
  hastnull_ = head->find(cat("TNULL",i)) ? 1:0;
  tnull_ = head->getInteger(cat("TNULL",i), 0);

  char* td = head->find(cat("TDMAX",i));
  char* tl = head->find(cat("TLMAX",i));
  char* ta = head->find(cat("TALEN",i));
  char* ax = head->find(cat("AXLEN",i));

  // this provides backward compatibility

  if (td) {
    hastlmin_ = head->find(cat("TDMIN",i)) ? 1:0;
    hastlmax_ = 1;
    tlmin_ = head->getReal(cat("TDMIN",i), 0);
    tlmax_ = head->getReal(cat("TDMAX",i), 0);
  }
  else if (tl) {
    hastlmin_ = head->find(cat("TLMIN",i)) ? 1:0;
    hastlmax_ = 1;
    tlmin_ = head->getReal(cat("TLMIN",i), 0);
    tlmax_ = head->getReal(cat("TLMAX",i), 0);
  }
  else if (ta) {
    hastlmin_ = 0;
    hastlmax_ = 1;
    tlmin_ = 1;
    tlmax_ = head->getReal(cat("TALEN",i), 0);
  }
  else if (ax) {
    hastlmin_ = 0;
    hastlmax_ = 1;
    tlmin_ = 1;
    tlmax_ = head->getReal(cat("AXLEN",i), 0);
  }
  else {
    hastlmin_ = 0;
    hastlmax_ = 0;
    tlmin_ = 0;
    tlmax_ = 0;
  }

  // now, make sure they are valid
  if (tlmin_>tlmax_) {
    hastlmin_ = 0;
    hastlmax_ = 0;
    tlmin_ = 0;
    tlmax_ = 0;
  }

  min_ = -DBL_MAX;
  max_ = DBL_MAX;
}

FitsColumn::~FitsColumn()
{
  if (tform_)
    delete [] tform_;
  if (tunit_)
    delete [] tunit_;
  if (ttype_)
    delete [] ttype_;
}

#if __GNUC__ >= 3
FitsAsciiColumn::FitsAsciiColumn(FitsHead* head, int i, int offset)
  : FitsColumn(head, i, offset)
{
  prec_ = 0;
  int tbcol = head->getInteger(cat("TBCOL",i),0);
  if (tbcol)
    offset_ = tbcol-1;

  // parse TFORM

  if (tform_) {
    string x(tform_);
    istringstream str(x);
    str >> type_ >> width_;
    if (type_ == 'F' || type_ == 'E' || type_ == 'D') {
      char s;
      str >> s >> prec_;
    }
  }
}
#else
FitsAsciiColumn::FitsAsciiColumn(FitsHead* head, int i, int offset)
  : FitsColumn(head, i, offset)
{
  prec_ = 0;
  int tbcol = head->getInteger(cat("TBCOL",i),0);
  if (tbcol)
    offset_ = tbcol-1;

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    str >> type_ >> width_;
    if (type_ == 'F' || type_ == 'E' || type_ == 'D') {
      char s;
      str >> s >> prec_;
    }
  }
}
#endif

#if __GNUC__ >= 3
double FitsAsciiColumn::value(const char* ptr, int i)
{
  string x(ptr+offset_);
  istringstream str(x);
  double r;
  str >> r;

  return r;
}
#else
double FitsAsciiColumn::value(const char* ptr, int i)
{
  istrstream str(ptr+offset_);
  double r;
  str >> r;

  return r;
}
#endif

char* FitsAsciiColumn::str(const char* ptr, int i)
{
  strncpy(buf_, ptr+offset_, width_);
  buf_[width_] = '\0';
  return buf_;
}

Vector FitsAsciiColumn::dimension()
{
  if (hastlmin_ || hastlmax_)
    return Vector(tlmin_,tlmax_);
  else
    switch (type_) {
    case 'A':
      return Vector(0,UCHAR_MAX);
      break;
    case 'I':
      return Vector(INT_MIN,INT_MAX);
      break;
    case 'F':
    case 'E':
      return Vector(FLT_MIN,FLT_MAX);
      break;
    case 'D':
      return Vector(DBL_MIN,DBL_MAX);
      break;
    }
}

FitsBinColumn::FitsBinColumn(FitsHead* head, int i, int offset)
  : FitsColumn(head, i, offset)
{
  repeat_ = 1;
  tdisp_ = head->getString(cat("TDISP",i));
}

FitsBinColumn::~FitsBinColumn()
{
  if (tdisp_)
    delete [] tdisp_;
}

#if __GNUC__ >= 3
FitsBinColumnA::FitsBinColumnA(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  Awidth_ = 0;
  Aterm_ = 0;

  // parse TFORM

  if (tform_) {
    string x(tform_);
    istringstream str(x);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;

    char s;
    str >> s >> s >> s >> s >> s >> Awidth_ >> s >> oct >> Aterm_;
  }

  // and determine field width

  width_ = repeat_ * (Awidth_+1);
}
#else
FitsBinColumnA::FitsBinColumnA(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  Awidth_ = 0;
  Aterm_ = 0;

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;

    char s;
    str >> s >> s >> s >> s >> s >> Awidth_ >> s >> oct >> Aterm_;
  }

  // and determine field width

  width_ = repeat_ * (Awidth_+1);
}
#endif

#if __GNUC__ >= 3
double FitsBinColumnA::value(const char* ptr, int i)
{
  double r;
  string x(ptr+offset_);
  istringstream str(x);
  str >> r;

  return r;
}
#else
double FitsBinColumnA::value(const char* ptr, int i)
{
  double r;
  istrstream str(ptr+offset_, width_);
  str >> r;

  return r;
}
#endif

char* FitsBinColumnA::str(const char* ptr, int i)
{
  strncpy(buf_, ptr+offset_, width_);
  buf_[width_] = '\0';
  return buf_;
}

Vector FitsBinColumnA::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_,tlmax_) : 
    Vector(DBL_MIN,DBL_MAX);
}

#if __GNUC__ >= 3
FitsBinColumnB::FitsBinColumnB(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  byteswap = lsb();

  // parse TFORM

  if (tform_) {
    string x(tform_);
    istringstream str(x);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;
  }
}
#else
FitsBinColumnB::FitsBinColumnB(FitsHead* head, int i, int offset)
  : FitsBinColumn(head, i, offset)
{
  byteswap = lsb();

  // parse TFORM

  if (tform_) {
    istrstream str(tform_);
    if (isalpha(tform_[0]))
      str >> type_;
    else
      str >> repeat_ >> type_;
  }
}
#endif

FitsBinColumnBit::FitsBinColumnBit(FitsHead* head, int i, int off)
  : FitsBinColumnB(head, i, off)
{
  // and determine field width

  width_ = (repeat_+7)/8;
}

double FitsBinColumnBit::value(const char* ptr, int i)
{
  // not implemented
  return 0;
}

char* FitsBinColumnBit::str(const char* ptr, int i)
{
  // not implemented
  return NULL;
}

Vector FitsBinColumnBit::dimension()
{
  return Vector(0,repeat_);
}

template<class T> FitsBinColumnT<T>::FitsBinColumnT(FitsHead* head,
						    int i, int off)
  : FitsBinColumnB(head, i, off)
{
  // and determine field width

  width_ = repeat_ * size();
}

unsigned char FitsBinColumnT<unsigned char>::Tmin() {return 0;}
unsigned char FitsBinColumnT<unsigned char>::Tmax() {return UCHAR_MAX;}

short FitsBinColumnT<short>::Tmin() {return SHRT_MIN;}
short FitsBinColumnT<short>::Tmax() {return SHRT_MAX;}

unsigned short FitsBinColumnT<unsigned short>::Tmin() {return 0;}
unsigned short FitsBinColumnT<unsigned short>::Tmax() {return USHRT_MAX;}

int FitsBinColumnT<int>::Tmin() {return INT_MIN;}
int FitsBinColumnT<int>::Tmax() {return INT_MAX;}

unsigned int FitsBinColumnT<unsigned int>::Tmin() {return 0;}
unsigned int FitsBinColumnT<unsigned int>::Tmax() {return UINT_MAX;}

float FitsBinColumnT<float>::Tmin() {return -FLT_MAX;}
float FitsBinColumnT<float>::Tmax() {return FLT_MAX;}

double FitsBinColumnT<double>::Tmin() {return -DBL_MAX;}
double FitsBinColumnT<double>::Tmax() {return DBL_MAX;}

unsigned char FitsBinColumnT<unsigned char>::swap(unsigned char* ptr)
{
  return *ptr;
}

short FitsBinColumnT<short>::swap(short* ptr)
{
  short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

unsigned short FitsBinColumnT<unsigned short>::swap(unsigned short* ptr)
{
  unsigned short r;
  swap2((char*)ptr, (char*)&r);
  return r;
}

int FitsBinColumnT<int>::swap(int* ptr)
{
  int r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

unsigned int FitsBinColumnT<unsigned int>::swap(unsigned int* ptr)
{
  unsigned int r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

float FitsBinColumnT<float>::swap(float* ptr)
{
  float r;
  swap4((char*)ptr, (char*)&r);
  return r;
}

double FitsBinColumnT<double>::swap(double* ptr)
{
  double r;
  swap8((char*)ptr, (char*)&r);
  return r;
}

// We need this to insure proper alignment

template<class T> double FitsBinColumnT<T>::value(const char* ptr, int i)
{
  return !byteswap ? *((T*)(ptr+offset_)+i) : swap(((T*)(ptr+offset_))+i);
}

double FitsBinColumnT<short>::value(const char* ptr, int i)
{
  union {
    char c[2];
    short s;
  } u;
  memcpy(u.c,ptr+offset_+i*2,2);
  return !byteswap ? u.s : swap(&u.s);
}

double FitsBinColumnT<unsigned short>::value(const char* ptr, int i)
{
  union {
    char c[2];
    unsigned short s;
  } u;
  memcpy(u.c,ptr+offset_+i*2,2);
  return !byteswap ? u.s : swap(&u.s);
}

double FitsBinColumnT<int>::value(const char* ptr, int i)
{
  union {
    char c[4];
    int l;
  } u;
  memcpy(u.c,ptr+offset_+i*4,4);
  return !byteswap ? u.l : swap(&u.l);
}

double FitsBinColumnT<unsigned int>::value(const char* ptr, int i)
{
  union {
    char c[4];
    unsigned int l;
  } u;
  memcpy(u.c,ptr+offset_+i*4,4);
  return !byteswap ? u.l : swap(&u.l);
}

double FitsBinColumnT<float>::value(const char* ptr, int i)
{
  union {
    char c[4];
    float f;
  } u;
  memcpy(u.c,ptr+offset_+i*4,4);
  return !byteswap ? u.f : swap(&u.f);
}

double FitsBinColumnT<double>::value(const char* ptr, int i)
{
  union {
    char c[8];
    double d;
  } u;
  memcpy(u.c,ptr+offset_+i*8,8);
  return !byteswap ? u.d : swap(&u.d);
}

#if __GNUC__ >= 3
template<class T> char* FitsBinColumnT<T>::str(const char* ptr, int i)
{
  ostringstream ost;
  ost << value(ptr,i) << ends;
  return (char*)ost.str().c_str();
}
#else
template<class T> char* FitsBinColumnT<T>::str(const char* ptr, int i)
{
  ostrstream ost(buf_,64);
  ost << value(ptr,i) << ends;
  return buf_;
}
#endif

Vector FitsBinColumnT<unsigned char>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_-.5,tlmax_+.5) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<short>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_-.5,tlmax_+.5) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<unsigned short>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_-.5,tlmax_+.5) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<int>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_-.5,tlmax_+.5) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<unsigned int>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_-.5,tlmax_+.5) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<float>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_,tlmax_) 
    : Vector(Tmin(),Tmax());
}

Vector FitsBinColumnT<double>::dimension()
{
  return (hastlmin_ || hastlmax_) ? Vector(tlmin_,tlmax_) 
    : Vector(Tmin(),Tmax());
}

template class FitsBinColumnT<unsigned char>;
template class FitsBinColumnT<short>;
template class FitsBinColumnT<unsigned short>;
template class FitsBinColumnT<int>;
template class FitsBinColumnT<unsigned int>;
template class FitsBinColumnT<float>;
template class FitsBinColumnT<double>;
