#include <Xlib.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "nan.h"

#include "alloc.h"
#include "allocgz.h"
#include "channel.h"
#include "mmap.h"
#include "share.h"
#include "socket.h"
#include "socketgz.h"
#include "var.h"
#include "contour.h"

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) : FrameBase(i, c, item)
{
  fits = NULL;
  fitsCount = 0;
  displayMode = SINGLE;
  mosaicType = NOMOSAIC;
  contour = NULL;
  
  currentFits = NULL;
  channelFits = &fits;
  keyFits = &fits;
  currentCount = &fitsCount;
  currentScale = &frScale;
  currentMode = &displayMode;
  currentMosaic = &mosaicType;

  currentWCScdelt = &WCScdelt;
  currentWCSrot = &WCSrot;
  currentWCSorientation = &WCSorientation;
  currentWCSbb = &WCSbb;

  currentcontour = &contour;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;

  iisSaveRoot = 0;
  iisInteractive = 0;
}

Frame::~Frame()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

// Virtual Routines with base components

void Frame::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;

  FrameBase::reset();
}

void Frame::setHistCursor(const Vector& v)
{
  if (currentFits)
    currentFits->setHistCursor(cursor);
}

void Frame::updateBin(const Matrix& m)
{
  // always center on center of pixel
  cursor = imageCenter() + Vector(.5,.5);

  // update bin
  crosshair *= m;
  markerbg();
  updateMarkerCoords(m);
  markerfg();
  updateMarkerCoords(m);

  // update contours
  // main contour
  if (hasContour())
    (*currentcontour)->updateCoords(m);

  // aux contours
  Contour* ptr=auxcontours.head();
  while (ptr) {
    ptr->updateCoords(m);
    ptr=ptr->next();
  }

  // finish bin
  setCurrentFits(*channelFits);
  int bd = (*channelFits)->binDepth();
  if (bd > 1) {
    FitsImage* ptr = (*channelFits)->next();
    while (ptr) {
      FitsImage* tmp = ptr->next();
      delete ptr;
      ptr = tmp;
    }
    *currentCount = 1;
    *currentMode = SINGLE;
    slice = 1;

    ptr = *channelFits;
    for (int i=1; i<bd; i++) {
      FitsImage* next = NULL;
      next = new FitsImageFitsNextHist(this, *channelFits, ptr->getImgFile());
      if (next && next->isValid()) {
	ptr->setNext(next);
	ptr = next;
	(*currentCount)++;
      }
      else {
	if (next)
	  delete next;
	break;
      }
    }

  }

  updateBinFileNames();

  updateClip();
  initHistogram();
  updateHistogram();

  align();
  update(MATRIX); // because we have changed zoom
}

BBox Frame::imageBBox()
{
  if (fits) {
    switch (mosaicType) {
    case NOMOSAIC:
      return BBox(0,0,fits->width(),fits->height());
    case IRAF:
      return fits->getIRAFbb();
    case WCSMOSAIC:
    case WFPC2:
      return *currentWCSbb;
    }
  }
  else
    return BBox();
}

void Frame::unloadFits()
{
  // clean up from iis if needed
  if (isIIS()) {
    frScale.setClipScope(FrScale::LOCAL);
    frScale.setClipMode(FrScale::MINMAX);
    frScale.setMMMode(FrScale::SAMPLE);

    setColorScale(FrScale::LINEARSCALE);
  }

  // now, delete any fitsimages
  while (fits) {
    FitsImage* tmp = fits->next();
    delete fits;
    fits = tmp;
  }
  fits = NULL;

  // delete any contour
  if (contour)
    delete contour;
  contour = NULL;

  FrameBase::unloadFits();
}

void Frame::align()
{
  if (wcsAlignment && hasWCS(wcsSystem)) {
    // set FrameRGB::align()
    //    wcsZoom = 1;
    wcsRotation = currentFits->getWCSRotation(wcsSystem,wcsSky,0);
    wcsOrientation = currentFits->getWCSOrientation(wcsSystem,wcsSky,0);
  }
  else {
    wcsZoom = 1;
    wcsRotation = 0;
    wcsOrientation = NORMAL;
  }

  switch (wcsOrientation) {
  case NORMAL:
    wcsOrientationMatrix.identity();
    break;
  case XX:
    wcsOrientationMatrix = FlipX();
    break;
  case YY:
    wcsOrientationMatrix = FlipY();
    break;
  case XY:
    wcsOrientationMatrix = FlipXY();
    break;
  }
}

void Frame::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  FitsImage* ptr = fits;
  Matrix rgb; // alway identity
  while (ptr) {
    ptr->updateMatrices(mosaicType, rgb, refToUser, userToWidget, 
			widgetToCanvas, userToPanner);
    ptr = ptr->next();
  }
}

void Frame::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMagnifierMatrices(refToMagnifier);
    ptr = ptr->next();
  }
}

// Virtual Routines

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

int Frame::isIIS() 
{
  return currentFits && currentFits->isIIS();
}

void Frame::psLevel1(PSColorSpace mode, int width, int height, float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel1Head(mode, width, height);

  switch (*currentMode) {
  case SINGLE:
    psLevel1Single(mode, width, height, scale);
    break;
  case MOSAIC:
    if (mosaicFast && *currentMosaic == IRAF)
      psLevel1MosaicFast(mode, width, height, scale);
    else
      psLevel1Mosaic(mode, width, height, scale);
    break;
  }
}

void Frame::psLevel2(PSColorSpace mode, int width, int height, float scale)
{
  // we need a colorScale before we can render
  if (!colorScale)
    return;

  psLevel2Head(mode, width, height);

  switch (*currentMode) {
  case SINGLE:
    psLevel2Single(mode, width, height, scale);
    break;
  case MOSAIC:
    if (mosaicFast && *currentMosaic == IRAF)
      psLevel2MosaicFast(mode, width, height, scale);
    else
      psLevel2Mosaic(mode, width, height, scale);
    break;
  }
}

#if __GNUC__ >= 3
void Frame::psLevel1Single(PSColorSpace mode, int width, int height, 
			   float scale)
{
  if (!currentFits)
    return;

  NoCompressAsciiHex filter;
  currentFits->updatePS(psMatrix(scale, width, height));
  double* m = currentFits->getmPSToData();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (int j=0; j<height; j++) {
    ostringstream str;

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      double x = i*m[0] + j*m[3] + m[6];
      double y = i*m[1] + j*m[4] + m[7];

      if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	double value = currentFits->getValueDouble(((int)y)*srcw + (int)x);

	if (!isNaNd(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
#else
void Frame::psLevel1Single(PSColorSpace mode, int width, int height, 
			   float scale)
{
  if (!currentFits)
    return;

  NoCompressAsciiHex filter;

  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + 
  // endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char* buf = new char[size];

  currentFits->updatePS(psMatrix(scale, width, height));

  double* m = currentFits->getmPSToData();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (int j=0; j<height; j++) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      double x = i*m[0] + j*m[3] + m[6];
      double y = i*m[1] + j*m[4] + m[7];

      if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	double value = currentFits->getValueDouble(((int)y)*srcw + (int)x);

	if (!isNaNd(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
  delete buf;
}
#endif

#if __GNUC__ >= 3
void Frame::psLevel1Mosaic(PSColorSpace mode, int width, int height,
			   float scale)
{
  NoCompressAsciiHex filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();
  FitsImage* ptr = fits;
  Matrix mm = psMatrix(scale, width, height);
  while (ptr) {
    ptr->updatePS(mm);
    ptr = ptr->next();
  }

  for (int j=0; j<height; j++) {
    ostringstream str;

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      ptr = fits;
      while (ptr) {
	double* m = ptr->getmPSToData();
	double x = i*m[0] + j*m[3] + m[6];
	double y = i*m[1] + j*m[4] + m[7];

	int* params = ptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  double ll = ptr->getLowDouble();
	  double hh = ptr->getHighDouble();
	  double diff = hh - ll;

	  if (!isNaNd(value))
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }

	  break;
	}
	ptr = ptr->next();
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
#else
void Frame::psLevel1Mosaic(PSColorSpace mode, int width, int height,
			   float scale)
{
  NoCompressAsciiHex filter;

  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + 
  // endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char* buf = new char[size];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  FitsImage* ptr = fits;
  Matrix mm = psMatrix(scale, width, height);
  while (ptr) {
    ptr->updatePS(mm);
    ptr = ptr->next();
  }

  for (int j=0; j<height; j++) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      ptr = fits;
      while (ptr) {
	double* m = ptr->getmPSToData();
	double x = i*m[0] + j*m[3] + m[6];
	double y = i*m[1] + j*m[4] + m[7];

	int* params = ptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  double ll = ptr->getLowDouble();
	  double hh = ptr->getHighDouble();
	  double diff = hh - ll;

	  if (!isNaNd(value))
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }

	  break;
	}
	ptr = ptr->next();
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
  delete buf;
}
#endif

#if __GNUC__ >= 3
void Frame::psLevel1MosaicFast(PSColorSpace mode, int width, int height,
			       float scale)
{
  if (!fits)
    return;

  NoCompressAsciiHex filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();
  FitsImage* ptr = fits;
  Matrix mm = psMatrix(scale, width, height);
  while (ptr) {
    ptr->updatePS(mm);
    ptr = ptr->next();
  }

  ptr = NULL;
  double* m;

  int* params;
  int srcw;
  int xmin;
  int xmax;
  int ymin;
  int ymax;

  double ll;
  double hh;
  double diff;

  for (int j=0; j<height; j++) {
    ostringstream str;

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      double x,y;

      if (ptr) {
	x = i*m[0] + j*m[3] + m[6];
	y = i*m[1] + j*m[4] + m[7];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  if (!isNaNd(value))
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	}
	else
	  ptr = NULL;
      }

      if (!ptr) {
	ptr = fits;
	while (ptr) {
	  m = ptr->getmPSToData();
	  x = i*m[0] + j*m[3] + m[6];
	  y = i*m[1] + j*m[4] + m[7];

	  params = ptr->getDataParams(frScale.scanMode());
	  srcw = params[0];
	  xmin = params[1];
	  xmax = params[2];
	  ymin = params[3];
	  ymax = params[4];

	  if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	    double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	    ll = ptr->getLowDouble();
	    hh = ptr->getHighDouble();
	    diff = hh - ll;

	    if (!isNaNd(value))
	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }

	    break;
	  }
	  ptr = ptr->next();
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
#else
void Frame::psLevel1MosaicFast(PSColorSpace mode, int width, int height,
			       float scale)
{
  if (!fits)
    return;

  NoCompressAsciiHex filter;

  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + 
  // endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char* buf = new char[size];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  FitsImage* ptr = fits;
  Matrix mm = psMatrix(scale, width, height);
  while (ptr) {
    ptr->updatePS(mm);
    ptr = ptr->next();
  }

  ptr = NULL;
  double* m;

  int* params;
  int srcw;
  int xmin;
  int xmax;
  int ymin;
  int ymax;

  double ll;
  double hh;
  double diff;

  for (int j=0; j<height; j++) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      double x,y;

      if (ptr) {
	x = i*m[0] + j*m[3] + m[6];
	y = i*m[1] + j*m[4] + m[7];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  if (!isNaNd(value))
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	}
	else
	  ptr = NULL;
      }

      if (!ptr) {
	ptr = fits;
	while (ptr) {
	  m = ptr->getmPSToData();
	  x = i*m[0] + j*m[3] + m[6];
	  y = i*m[1] + j*m[4] + m[7];

	  params = ptr->getDataParams(frScale.scanMode());
	  srcw = params[0];
	  xmin = params[1];
	  xmax = params[2];
	  ymin = params[3];
	  ymax = params[4];

	  if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	    double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	    ll = ptr->getLowDouble();
	    hh = ptr->getHighDouble();
	    diff = hh - ll;

	    if (!isNaNd(value))
	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }

	    break;
	  }
	  ptr = ptr->next();
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
  delete buf;
}
#endif

void Frame::buildXImage(XImage* xmap, int x0, int y0, int x1, int y1,
			double* (FitsImage::*getMatrix)())
{
  switch (*currentMode) {
  case SINGLE:
    fillXImage(xmap, x0, y0, x1, y1, getMatrix);
    break;
  case MOSAIC:
    if (mosaicFast && *currentMosaic == IRAF)
      fillXImageMosaicFast(xmap, x0, y0, x1, y1, getMatrix);
    else
      fillXImageMosaic(xmap, x0, y0, x1, y1, getMatrix);
    break;
  }
}

// Fits Commands

void Frame::getFitsDepthCmd()
{
  if (fits && displayMode==SINGLE)
    printInteger(fitsCount);
  else
    printInteger(1);
}

void Frame::hasFitsCubeCmd()
{
  if (fits && displayMode==SINGLE && fitsCount>1)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::sliceCmd(int s)
{
  if (fits && displayMode==SINGLE && (s>0) && (s<=fitsCount)) {
    slice = s;
    currentFits = fits;
    for (int i=1; i<s; i++)
      currentFits = currentFits->next();
  }

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}

// Colormap Commands

#if __GNUC__ >= 3
void Frame::getColormapCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
#else
void Frame::getColormapCmd()
{
  char buf[64];
  ostrstream str(buf,64);
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;

  Tcl_AppendResult(interp, buf, NULL);
}
#endif

void Frame::getTypeCmd()
{
  Tcl_AppendResult(interp, "base", NULL);
}

// Pan Zoom Rotate Orient Commands

void Frame::getWCSZoomCmd(CoordSystem sys, Precision p)
{
  if (hasWCS(sys)) {
    Vector v = currentFits->getWCScdelt(sys);
    printDouble(v[0], p);
  }
  else
    Tcl_AppendResult(interp, "1", NULL);
}

void Frame::wcsZoomCmd(CoordSystem sys, double a)
{
  if (hasWCS(sys)) {
    Vector b = currentFits->getWCScdelt(sys);
    wcsZoom = fabs(b[0]/a);
  }
  else
    wcsZoom = 1;

  update(MATRIX);
}

// IIS Commands

void Frame::iisCmd(int width, int height)
{
  unloadAllFits();

  // force clip mode to user between 1 to 200

  frScale.setClipMode(FrScale::USERCLIP);
  frScale.setMMMode(FrScale::SAMPLE);
  frScale.setMMIncr(1);
  frScale.setULow(IISMIN);
  frScale.setUHigh(IISMAX);
  
  setColorScale(FrScale::IISSCALE);

  FitsImage* img = new FitsImageIIS(this, width, height);

  load(ALLOC,"",img);
}

void Frame::iisEraseCmd()
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisErase();
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (currentFits)
    memcpy(ptr, ((FitsImageIIS*)currentFits)->iisGet(x,y), dx*dy);
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& m, const Vector& z, int zt)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisWCS(m, z, zt);
}

void Frame::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}


