#include "framergb.h"
#include "colorscalergb.h"
#include "fitsimage.h"
#include "ps.h"
#include "util.h"
#include "nan.h"
#include "contour.h"

// Frame Member Functions

FrameRGB::FrameRGB(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i, c, item)
{
  channel = 0;
  rgbSystem = WCS;

  for (int ii=0; ii<3; ii++) {
    view[ii] = 1;

    fits[ii] = NULL;
    cfits[ii] = NULL;
    fitsCount[ii] = 0;

    bias[ii] = .5;
    contrast[ii] = 1.0;

    colorScale[ii] = NULL;
    
    displayMode[ii] = SINGLE;
    mosaicType[ii] = NOMOSAIC;

    WCSrot[ii] = 0;
    WCSorientation[ii] = NORMAL;

    contour[ii] = NULL;
  }

  colorCount = 0;
  colorCells = NULL;

  setCurrentFits(NULL);
  keyFits = NULL;
  channelFits = &fits[channel];
  currentCount = &fitsCount[channel];
  currentScale = &frScale[channel];
  currentMode = &displayMode[channel];
  currentMosaic = &mosaicType[channel];

  currentWCScdelt = &WCScdelt[channel];
  currentWCSrot = &WCSrot[channel];
  currentWCSorientation = &WCSorientation[channel];
  currentWCSbb = &WCSbb[channel];

  currentcontour = &contour[channel];

  isRGBCube = 0;
}

FrameRGB::~FrameRGB()
{
  for (int ii=0; ii<3; ii++) {
    if (colorScale[ii])
      delete colorScale[ii];
  }

  if (colorCells)
    delete colorCells;
}

void FrameRGB::reset()
{
  for (int ii=0; ii<3; ii++) {
    bias[ii] = 0.5;
    contrast[ii] = 1.0;
  }

  isRGBCube = 0;
  FrameBase::reset();
}

void FrameRGB::setCurrentFits(FitsImage* f)
{
  currentFits = f;
  cfits[channel] = f;
}

void FrameRGB::setHistCursor(const Vector& v)
{
  for (int ii=0; ii<3; ii++)
    if (fits[ii])
      fits[ii]->setHistCursor(cursor);
}

void FrameRGB::updateBin(const Matrix& m)
{
  // always center on center of pixel
  cursor = imageCenter() + Vector(.5,.5);

  if (hasKeyFits() && (*keyFits == currentFits)) {
    // update markers,crosshair,aux contours
    crosshair *= m;
    markerbg();
    updateMarkerCoords(m);
    markerfg();
    updateMarkerCoords(m);

    // aux contours
    Contour* ptr=auxcontours.head();
    while (ptr) {
      ptr->updateCoords(m);
      ptr=ptr->next();
    }
  }

  FitsImage* currentSave = currentFits;

  for (int ii=0; ii<3; ii++) 
    if (fits[ii] && fits[ii]->isHist()) {
      // update bin

      if (fits[ii] == currentSave && contour[ii])
	contour[ii]->updateCoords(m);

      currentFits = fits[ii];
      cfits[ii] = fits[ii];
      currentScale = &frScale[ii];
      channelFits = &fits[ii];
      currentCount = &fitsCount[ii];
      currentMode = &displayMode[ii];
      currentMosaic = &mosaicType[ii];

      // finish bin
      int bd = (*channelFits)->binDepth();
      if (bd > 1) {
	FitsImage* ptr = (*channelFits)->next();
	while (ptr) {
	  FitsImage* tmp = ptr->next();
	  delete ptr;
	  ptr = tmp;
	}
	*currentCount = 1;
	*currentMode = SINGLE;
	slice = 1;

	ptr = *channelFits;
	for (int i=1; i<bd; i++) {
	  FitsImage* next = NULL;
	  next = 
	    new FitsImageFitsNextHist(this,*channelFits, ptr->getImgFile());
	  if (next && next->isValid()) {
	    ptr->setNext(next);
	    ptr = next;
	    (*currentCount)++;
	  }
	  else {
	    if (next)
	      delete next;
	    break;
	  }
	}

      }

      updateBinFileNames();

      updateClip();
      initHistogram();
      updateHistogram();
    }

  slice = 1;
  currentFits = fits[channel];
  cfits[channel] = fits[channel];
  currentScale = &frScale[channel];

  channelFits = &fits[channel];
  currentCount = &fitsCount[channel];
  currentMode = &displayMode[channel];
  currentMosaic = &mosaicType[channel];

  currentWCScdelt = &WCScdelt[channel];
  currentWCSrot = &WCSrot[channel];
  currentWCSorientation = &WCSorientation[channel];
  currentWCSbb = &WCSbb[channel];

  currentcontour = &contour[channel];

  align();
  update(MATRIX); // because we have changed zoom
}

BBox FrameRGB::imageBBox()
{
  // make sure we have the rgb matrices defined
  align();

  BBox r;
  int first=1;
  for (int ii=0; ii<3; ii++) {
    if (fits[ii]) {
      if (first) {
	switch (mosaicType[ii]) {
	case NOMOSAIC:
	  // no need for bounding since the key matrix should be identity
	  r = BBox(Vector(1,1), Vector(fits[ii]->width(), fits[ii]->height()));
	  break;
	case IRAF:
	  r = fits[ii]->getIRAFbb();
	  break;
	case WCSMOSAIC:
	case WFPC2:
	  r = WCSbb[ii];
	  break;
	}
	first = 0;
      } 
      else {
	switch (mosaicType[ii]) {
	case NOMOSAIC:
	  r.bound(Vector(0,0) * rgb[ii]);
	  r.bound(Vector(fits[ii]->width(),0) * rgb[ii]);
	  r.bound(Vector(fits[ii]->width(),fits[ii]->height()) * rgb[ii]);
	  r.bound(Vector(0,fits[ii]->height()) * rgb[ii]);
	  break;
	case IRAF:
	  r.bound(fits[ii]->getIRAFbb());
	  break;
	case WCSMOSAIC:
	case WFPC2:
	  r.bound(WCSbb[ii]);
	  break;
	}
      }
    }
  }
  return r;
}

void FrameRGB::unloadFits()
{
  while (fits[channel]) {
    FitsImage* tmp = fits[channel]->next();
    delete fits[channel];
    fits[channel] = tmp;
  }
  fits[channel] = NULL;
  cfits[channel] = NULL;

  if (contour[channel])
    delete contour[channel];
  contour[channel] = NULL;

  rgb[channel].identity();

  keyFits = NULL;
  slice = 1;
  setCurrentFits(NULL);
  *currentCount = 0;
  *currentMode = SINGLE;
  *currentMosaic = NOMOSAIC;

  // delete histograms (equalization,blt)
  initHistogram();
}

void FrameRGB::unloadAllFits()
{
  for (int ii=0; ii<3; ii++) {
    while (fits[ii]) {
      FitsImage* tmp = fits[ii]->next();
      delete fits[ii];
      fits[ii] = tmp;
    }
    fits[ii] = NULL;
    cfits[ii] = NULL;
    fitsCount[ii] = 0;
    displayMode[ii] = SINGLE;
    mosaicType[ii] = NOMOSAIC;

    WCScdelt[ii] = Vector();
    WCSrot[ii] = 0;
    WCSorientation[ii] = NORMAL;
    WCSbb[ii] = BBox();

    if (contour[ii])
      delete contour[ii];
    contour[ii] = NULL;

    frScale[ii].deleteHistogramX();
    frScale[ii].deleteHistogramY();
    frScale[ii].deleteHistequ();

    rgb[ii].identity();

    if (frScale[ii].colorScaleType() == FrScale::HISTEQUSCALE)
      updateColorScale();
  }

  keyFits = NULL;
  FrameBase::unloadFits();
}

void FrameRGB::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  for (int ii=0; ii<3; ii++) {
    FitsImage* ptr = fits[ii];
    while (ptr) {
      ptr->updateMatrices(mosaicType[ii], rgb[ii], refToUser, userToWidget,
			  widgetToCanvas, userToPanner);
      ptr = ptr->next();
    }
  }
}

void FrameRGB::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  for (int ii=0; ii<3; ii++) {
    FitsImage* ptr = fits[ii];
    while (ptr) {
      ptr->updateMagnifierMatrices(refToMagnifier);
      ptr = ptr->next();
    }
  }
}

void FrameRGB::updateColorCells(unsigned char* cells, int cnt)
{
  if (DebugRGB) 
    cerr << "updateColorCells" << endl;

  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }

  memcpy(colorCells, cells, cnt*3);
}

void FrameRGB::updateColorScale()
{
  // we need colors before we can construct a scale
  if (!colorCells)
    return;

  if (DebugRGB) 
    cerr << "updateColorScale" << endl;

  // since we need to know about bytes per pixel and byte order, 
  // we may need to wait til we have an XImage available and try again
  //  if (!baseXImage)
  //    return;

  for (int ii=0; ii<3; ii++) {
    if (colorScale[ii])
      delete colorScale[ii];

    switch (frScale[ii].colorScaleType()) {
    case FrScale::LINEARSCALE:
      colorScale[ii] = 
	new LinearScaleRGB(ii, colorCount, colorCells, colorCount);
      break;
    case FrScale::LOGSCALE:
      colorScale[ii] =
	new LogScaleRGB(ii, SCALESIZE, colorCells, colorCount);
      break;
    case FrScale::SQUAREDSCALE:
      colorScale[ii] =
	new SquaredScaleRGB(ii, SCALESIZE, colorCells, colorCount);
      break;
    case FrScale::SQRTSCALE:
      colorScale[ii] =
	new SqrtScaleRGB(ii, SCALESIZE, colorCells, colorCount);
      break;
    case FrScale::HISTEQUSCALE:
      calcHistEqu();
      colorScale[ii] = 
	new HistEquScaleRGB(ii, SCALESIZE, colorCells, colorCount, 
			    frScale[ii].histequ(), HISTEQUSIZE);
      break;
    }
  }
}

void FrameRGB::align()
{
  // set keyfits
  keyFits = NULL;
  for (int ii=0; ii<3; ii++)
    if (fits[ii]) {
	keyFits = &fits[ii]; // first available becomes the key
	break;
    }
  
  // wcs align
  if (wcsAlignment && hasKeyFits() && (*keyFits)->hasWCS(wcsSystem)) {
    // this causes problems because imageBBox() call align()
    //    wcsZoom = 1;
    wcsRotation = (*keyFits)->getWCSRotation(wcsSystem,wcsSky,1);
    wcsOrientation = (*keyFits)->getWCSOrientation(wcsSystem,wcsSky,1);
  }
  else {
    wcsZoom = 1;
    wcsRotation = 0;
    wcsOrientation = NORMAL;
  }

  switch (wcsOrientation) {
  case NORMAL:
    wcsOrientationMatrix.identity();
    break;
  case XX:
    wcsOrientationMatrix = FlipX();
    break;
  case YY:
    wcsOrientationMatrix = FlipY();
    break;
  case XY:
    wcsOrientationMatrix = FlipXY();
    break;
  }

  // image,pysical,amplifier,detector are ok, check for wcs
  if (rgbSystem >= WCS) {
    for (int ii=0; ii<3; ii++) {
      if (fits[ii] && !fits[ii]->hasWCS(rgbSystem)) {
	// ok, don't have requested coordinate system
	// down grade to image
	rgbSystem = IMAGE;
	break;
      }
    }
  }

  // rgb align
  for (int ii=0; ii<3; ii++) {
    rgb[ii].identity();

    if (fits[ii] && fits[ii] != *keyFits) {
      switch (rgbSystem) {
      case IMAGE:
	// nothing to do here
	break;
      case PHYSICAL:
	rgbAlignPhysical(ii);
	break;
      case AMPLIFIER:
	rgbAlignAmplifier(ii);
	break;
      case DETECTOR:
	rgbAlignDetector(ii);
	break;
      default:
	rgbAlignWCS(ii);
	break;
      }
    }
  }
}   

void FrameRGB::rgbAlignPhysical(int ii)
{
  if (hasKeyFits())
    rgb[ii] = fits[ii]->getImageToPhysical()*(*keyFits)->getPhysicalToImage();

  if (DebugRGB) 
    cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
}

void FrameRGB::rgbAlignAmplifier(int ii)
{
  if (hasKeyFits())
    rgb[ii] = fits[ii]->getImageToAmplifier() * 
      (*keyFits)->getAmplifierToImage();

  if (DebugRGB) 
    cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
}

void FrameRGB::rgbAlignDetector(int ii)
{
  if (hasKeyFits())
    rgb[ii] = fits[ii]->getImageToDetector() * 
      (*keyFits)->getDetectorToImage();

  if (DebugRGB) 
    cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
}

void FrameRGB::rgbAlignWCS(int ii)
{
  if (hasKeyFits() && (*keyFits)->hasWCSEqu(rgbSystem)) {
    // rotation
    double rot = (*keyFits)->getWCSRotation(rgbSystem,FK5,0) - 
      fits[ii]->getWCSRotation(rgbSystem,FK5,0);

    // zoom
    double zz = fabs(fits[ii]->getWCScdelt(rgbSystem)[0] / 
		     (*keyFits)->getWCScdelt(rgbSystem)[0]);

    // orientation
    Matrix flip;
    Orientation fitsOrientation = 
      fits[ii]->getWCSOrientation(rgbSystem,FK5,0);

    switch ((*keyFits)->getWCSOrientation(rgbSystem,FK5,0)) {
    case NORMAL:
      switch (fitsOrientation) {
      case NORMAL:
	// nothing
	break;
      case XX:
	flip = FlipX();
	break;
      case YY:
	flip = FlipY();
	break;
      case XY:
	flip = FlipXY();
	break;
      }
      break;
    case XX:
      switch (fitsOrientation) {
      case NORMAL:
	flip = FlipX();
	break;
      case XX:
	// nothing
	break;
      case YY:
	flip = FlipXY();
	break;
      case XY:
	flip = FlipY();
	break;
      }
      break;
    case YY:
      switch (fitsOrientation) {
      case NORMAL:
	flip = FlipY();
	break;
      case XX:
	flip = FlipXY();
	break;
      case YY:
	// nothing
	break;
      case XY:
	flip = FlipX();
	break;
      }
      break;
    case XY:
      switch (fitsOrientation) {
      case NORMAL:
	flip = FlipXY();
	break;
      case XX:
	flip = FlipY();
	break;
      case YY:
	flip = FlipX();
	break;
      case XY:
	// nothing
	break;
      }
      break;
    }

    // origin
    Vector aa = fits[ii]->getWCScrpix(rgbSystem);
    Vector bb = fits[ii]->pix2wcs(aa,rgbSystem,FK5);
    Vector org = (*keyFits)->wcs2pix(bb,rgbSystem,FK5);

    // calculate matrix
    rgb[ii] = 
      Translate(-fits[ii]->getWCScrpix(rgbSystem)) *
      Rotate(rot) *
      flip *
      Scale(zz) *
      Translate(org);
  }

  if (DebugRGB) 
    cerr << "rgb[" << ii << "] " << rgb[ii] << endl;
}

// Fits Commands

void FrameRGB::getFitsDepthCmd()
{
  if (isRGBCube) {
    printInteger(1);
    return;
  }

  if (fits[channel] && displayMode[channel]==SINGLE)
    printInteger(fitsCount[channel]);
  else
    printInteger(1);
}

void FrameRGB::hasFitsCubeCmd()
{
  if (isRGBCube) {
    Tcl_AppendResult(interp, "0", NULL);
    return;
  }
    
  if (fits[channel] && displayMode[channel]==SINGLE && fitsCount[channel]>1)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameRGB::sliceCmd(int s)
{

  for (int ii=0; ii<3; ii++) {
    cfits[ii] = fits[ii];
    if (fits[ii] && displayMode[ii]==SINGLE && (s>0) && (s<=fitsCount[ii])) {
      for (int i=1; i<s; i++)
	cfits[ii] = cfits[ii]->next();
    }
  }

  currentFits = cfits[channel];
  slice = s;

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}

// Colormap Commands

#if __GNUC__ >= 3
void FrameRGB::getColormapCmd()
{
  ostringstream str;

  str << setiosflags(ios::fixed);
  for (int ii=0; ii<3; ii++)
    str << bias[ii] << ' ';
  for (int ii=0; ii<3; ii++)
    str << contrast[ii] << ' ';
  str << ends;

  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
#else
void FrameRGB::getColormapCmd()
{
  char buf[128];
  ostrstream str(buf,128);

  str << setiosflags(ios::fixed);
  for (int ii=0; ii<3; ii++)
    str << bias[ii] << ' ';
  for (int ii=0; ii<3; ii++)
    str << contrast[ii] << ' ';
  str << ends;

  Tcl_AppendResult(interp, buf, NULL);
}
#endif

void FrameRGB::getTypeCmd()
{
  Tcl_AppendResult(interp, "rgb", NULL);
}

// Pan Zoom Rotate Orient Commands

void FrameRGB::getWCSZoomCmd(CoordSystem sys, Precision p)
{
  if (hasKeyFits() && (*keyFits)->hasWCS(wcsSystem)) {
    Vector v = (*keyFits)->getWCScdelt(sys);
    printDouble(v[0], p);
  }
  else
    Tcl_AppendResult(interp, "1", NULL);
}

void FrameRGB::wcsZoomCmd(CoordSystem sys, double a)
{
  if (hasKeyFits() && (*keyFits)->hasWCS(wcsSystem)) {
    Vector b = (*keyFits)->getWCScdelt(sys);
    wcsZoom = fabs(b[0]/a);
  }
  else
    wcsZoom = 1;

  update(MATRIX);
}

// RGB Commands

void FrameRGB::getRGBChannelCmd()
{
  switch (channel) {
  case 0:
    Tcl_AppendResult(interp, "red", NULL);
    return;
  case 1:
    Tcl_AppendResult(interp, "green", NULL);
    return;
  case 2:
    Tcl_AppendResult(interp, "blue", NULL);
    return;
  }
}

void FrameRGB::setRGBChannelCmd(const char* c)
{
  if (!strncmp(c,"red",3))
    channel = 0;
  else if (!strncmp(c,"gre",3))
    channel = 1;
  else if (!strncmp(c,"blu",3))
    channel = 2;
  else
    channel = 0;

  channelFits = &fits[channel];
  currentFits = cfits[channel];
  currentCount = &fitsCount[channel];
  currentScale = &frScale[channel];
  currentMode = &displayMode[channel];
  currentMosaic = &mosaicType[channel];

  currentWCScdelt = &WCScdelt[channel];
  currentWCSrot = &WCSrot[channel];
  currentWCSorientation = &WCSorientation[channel];
  currentWCSbb = &WCSbb[channel];

  currentcontour = &contour[channel];

  // execute any update callbacks
  updateCBMarkers();

 // always update
  update(BASE);
}

void FrameRGB::getRGBViewCmd()
{
  for (int ii=0; ii<3; ii++)
    Tcl_AppendElement(interp, view[ii] ? "1" : "0");
}

void FrameRGB::setRGBViewCmd(int r, int g, int b)
{
  view[0] = r ? 1 : 0;
  view[1] = g ? 1 : 0;
  view[2] = b ? 1 : 0;

  update(BASE); // always update
}

void FrameRGB::getRGBSystemCmd()
{
  printCoordSystem(rgbSystem);
}

void FrameRGB::setRGBSystemCmd(CoordSystem sys)
{
  rgbSystem = sys;

  // save current matrix
  Matrix old[3];
  for (int ii=0; ii<3; ii++)
    old[ii] = rgb[ii];

  align();

  // fix any contours
  for(int ii=0; ii<3; ii++)
    if (contour[ii]) {
      Matrix mm = old[ii].invert() * rgb[ii];
      contour[ii]->updateCoords(mm);
    }

  update(MATRIX);
}

// Load Commands
// RGB Cube

void FrameRGB::loadRGBCubeAllocCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsAlloc(this, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOC,fn,img);
}

void FrameRGB::loadRGBCubeAllocGZCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsAllocGZ(this, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOCGZ,fn,img);
}

void FrameRGB::loadRGBCubeChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsChannel(this, interp, ch, fn, 
					    FitsFile::NOFLUSH);
  loadRGBCube(CHANNEL,fn,img);
}

void FrameRGB::loadRGBCubeMMapCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsMMap(this, fn);
  setScanModeIncr(lm);
  loadRGBCube(MMAP,fn,img);
}

void FrameRGB::loadRGBCubeMMapIncrCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsMMapIncr(this, fn);
  setScanModeIncr(lm);
  loadRGBCube(MMAPINCR,fn,img);
}

void FrameRGB::loadRGBCubeShareCmd(ShmType type, int id, const char* fn,
				   LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsShare(this, type, id, fn);
  setScanModeIncr(lm);
  loadRGBCube(SHARE,fn,img);
}

void FrameRGB::loadRGBCubeSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKET,fn,img);
}

void FrameRGB::loadRGBCubeSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKETGZ,fn,img);
}

void FrameRGB::loadRGBCubeVarCmd(const char* ch, const char* fn,
				     LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageFitsVar(this, interp, ch, fn);
  setScanModeIncr(lm);
  loadRGBCube(VAR,fn,img);
}

// RGB Slave

void FrameRGB::loadRGBCubeSlaveCmd(const char* fn, FitsFile* f)
{
  unloadAllFits();
  FitsImage* img = new FitsImageSlave(this,fn,f);
  loadRGBCube(MMAPINCR,fn,img);
}

// RGB Image

void FrameRGB::loadRGBImageAllocCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicAlloc(this, fn, FitsFile::NOFLUSH);
  loadRGBImage(ALLOC,fn,img);
}

void FrameRGB::loadRGBImageAllocGZCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicAllocGZ(this, fn, FitsFile::NOFLUSH);
  loadRGBImage(ALLOCGZ,fn,img);
}

void FrameRGB::loadRGBImageChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicChannel(this, interp, ch, fn, 
					      FitsFile::NOFLUSH);
  loadRGBCube(CHANNEL,fn,img);
}

void FrameRGB::loadRGBImageMMapCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicMMap(this, fn);
  setScanModeIncr(lm);
  loadRGBImage(MMAP,fn,img);
}

void FrameRGB::loadRGBImageMMapIncrCmd(const char* fn, LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicMMapIncr(this, fn);
  setScanModeIncr(lm);
  loadRGBImage(MMAPINCR,fn,img);
}

void FrameRGB::loadRGBImageShareCmd(ShmType type, int id, const char* fn,
				    LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicShare(this, type, id, fn);
  setScanModeIncr(lm);
  loadRGBImage(SHARE,fn,img);
}

void FrameRGB::loadRGBImageSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBImage(SOCKET,fn,img);
}

void FrameRGB::loadRGBImageSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBImage(SOCKETGZ,fn,img);
}

void FrameRGB::loadRGBImageVarCmd(const char* ch, const char* fn,
				      LoadMethod lm)
{
  unloadAllFits();
  FitsImage* img = new FitsImageMosaicVar(this, interp, ch, fn);
  setScanModeIncr(lm);
  loadRGBImage(VAR,fn,img);
}

// Array RGB Cube

void FrameRGB::loadArrRGBCubeAllocCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrAlloc(this, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOC,fn,img);
}

void FrameRGB::loadArrRGBCubeAllocGZCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrAllocGZ(this, fn, FitsFile::NOFLUSH);
  loadRGBCube(ALLOCGZ,fn,img);
}

void FrameRGB::loadArrRGBCubeChannelCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrChannel(this, interp, ch, fn, 
					   FitsFile::NOFLUSH);
  loadRGBCube(CHANNEL,fn,img);
}

void FrameRGB::loadArrRGBCubeMMapCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrMMap(this, fn);
  loadRGBCube(MMAP,fn,img);
}

void FrameRGB::loadArrRGBCubeMMapIncrCmd(const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrMMapIncr(this, fn);
  loadRGBCube(MMAPINCR,fn,img);
}

void FrameRGB::loadArrRGBCubeShareCmd(ShmType type, int id, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrShare(this, type, id, fn);
  loadRGBCube(SHARE,fn,img);
}

void FrameRGB::loadArrRGBCubeSocketCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrSocket(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKET,fn,img);
}

void FrameRGB::loadArrRGBCubeSocketGZCmd(int s, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrSocketGZ(this, s, fn, FitsFile::FLUSH);
  loadRGBCube(SOCKETGZ,fn,img);
}

void FrameRGB::loadArrRGBCubeVarCmd(const char* ch, const char* fn)
{
  unloadAllFits();
  FitsImage* img = new FitsImageArrVar(this, interp, ch, fn);
  loadRGBCube(VAR,fn,img);
}

// loadRGB

void FrameRGB::loadRGBCube(MemType which, const char* fn, FitsImage* img)
{
  if (!img || !img->isValid() || !img->isImage() || (img->depth() != 3))
    goto error;

  fits[0] = img;

  switch (which) {
  case ALLOC:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextAlloc(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextAlloc(this, fn, fits[1]->getFitsFile());
    break;
  case ALLOCGZ:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextAllocGZ(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextAllocGZ(this, fn, fits[1]->getFitsFile());
    break;
  case CHANNEL:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextChannel(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextChannel(this, fn, fits[1]->getFitsFile());
    break;
  case MMAP:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextMMap(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextMMap(this, fn, fits[1]->getFitsFile());
    break;
  case MMAPINCR:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextMMapIncr(this, fn,fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextMMapIncr(this, fn,fits[1]->getFitsFile());
    break;
  case SHARE:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextShare(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextShare(this, fn, fits[1]->getFitsFile());
    break;
  case SOCKET:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSocket(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSocket(this, fn, fits[1]->getFitsFile());
    break;
  case SOCKETGZ:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextSocketGZ(this, fn,fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextSocketGZ(this, fn,fits[1]->getFitsFile());
    break;
  case VAR:
    if (fits[0] && fits[0]->isValid())
      fits[1] = new FitsImageFitsNextVar(this, fn, fits[0]->getFitsFile());
    if (fits[1] && fits[1]->isValid())
      fits[2] = new FitsImageFitsNextVar(this, fn, fits[1]->getFitsFile());
    break;
  }

  // is everything ok?
  if (fits[0] && fits[0]->isValid() && fits[0]->isImage() &&
      fits[1] && fits[1]->isValid() && fits[1]->isImage() &&
      fits[2] && fits[2]->isValid() && fits[2]->isImage()) {

    isRGBCube = 1;
    loadRGBFinish();
    return;
  }

 error:
  unloadAllFits();

  reset();

  updateClip();
  update(MATRIX);
    
  Tcl_AppendResult(interp, "Unable to load rgb cube file", NULL);
  result = TCL_ERROR;
  return;
}

void FrameRGB::loadRGBImage(MemType which, const char* fn, FitsImage* img)
{
  FitsImage* r = img;
  FitsImage* g = NULL;
  FitsImage* b = NULL;

  if (!img || !img->isValid() || !img->isImage())
    goto error;

  switch (which) {
  case ALLOC:
    if (r && r->isValid())
      g = new FitsImageMosaicNextAlloc(this, fn, r->getFitsFile(), 
				       FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextAlloc(this, fn, g->getFitsFile(), 
				       FitsFile::NOFLUSH);
    break;
  case ALLOCGZ:
    if (r && r->isValid())
      g = new FitsImageMosaicNextAllocGZ(this,fn,r->getFitsFile(), 
					 FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextAllocGZ(this,fn,g->getFitsFile(), 
					 FitsFile::NOFLUSH);
    break;
  case CHANNEL:
    if (r && r->isValid())
      g = new FitsImageMosaicNextChannel(this,fn,r->getFitsFile(), 
					 FitsFile::NOFLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextChannel(this,fn,g->getFitsFile(), 
					 FitsFile::NOFLUSH);
    break;
  case MMAP:
    if (r && r->isValid())
      g = new FitsImageMosaicNextMMap(this, fn, r->getFitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextMMap(this, fn, g->getFitsFile());
    break;
  case MMAPINCR:
    if (r && r->isValid())
      g = new FitsImageMosaicNextMMapIncr(this, fn, r->getFitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextMMapIncr(this, fn, g->getFitsFile());
    break;
  case SHARE:
    if (r && r->isValid())
      g = new FitsImageMosaicNextShare(this,fn, r->getFitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextShare(this,fn, g->getFitsFile());
    break;
  case SOCKET:
    if (r && r->isValid())
      g = new FitsImageMosaicNextSocket(this,fn,r->getFitsFile(), 
					FitsFile::FLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextSocket(this,fn,g->getFitsFile(), 
					FitsFile::FLUSH);
    break;
  case SOCKETGZ:
    if (r && r->isValid())
      g = new FitsImageMosaicNextSocketGZ(this,fn,r->getFitsFile(), 
					  FitsFile::FLUSH);
    if (g && g->isValid())
      b = new FitsImageMosaicNextSocketGZ(this,fn,g->getFitsFile(), 
					  FitsFile::FLUSH);
    break;
  case VAR:
    if (r && r->isValid())
      g = new FitsImageMosaicNextVar(this, fn, r->getFitsFile());
    if (g && g->isValid())
      b = new FitsImageMosaicNextVar(this, fn, g->getFitsFile());
    break;
  }

  // ok, figure out which is which channel
  fits[0] = fits[1] = fits[2] = NULL;

  if (r->findKeyword("EXTNAME")) {
    char* c = r->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = r;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = r;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = r;
  }
  else
    fits[0] = r;

  if (g->findKeyword("EXTNAME")) {
    char* c = g->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = g;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = g;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = g;
  }
  else
    fits[1] = g;

  if (b->findKeyword("EXTNAME")) {
    char* c = b->getKeyword("EXTNAME");
    if (!strncmp(c,"RED",3))
      fits[0] = b;
    else if (!strncmp(c,"GREEN",3))
      fits[1] = b;
    else if (!strncmp(c,"BLUE",3))
      fits[2] = b;
  }
  else
    fits[2] = b;

  // is everything ok?
  if (fits[0] && fits[0]->isValid() && fits[0]->isImage() &&
      fits[1] && fits[1]->isValid() && fits[1]->isImage() &&
      fits[2] && fits[2]->isValid() && fits[2]->isImage()) {

    loadRGBFinish();
    return;
  }

 error:
  unloadAllFits();

  reset();

  updateClip();
  update(MATRIX);
    
  Tcl_AppendResult(interp, "Unable to load rgb image file", NULL);
  result = TCL_ERROR;
  return;
}

void FrameRGB::loadRGBFinish()
{
  for (int ii=0; ii<3; ii++) {
    cfits[ii] = fits[ii];
    fitsCount[ii] = 1;
    displayMode[ii] = SINGLE;
    mosaicType[ii] = NOMOSAIC;
  }

  // update clip values

  channel = 0;
  slice = 1;
  for (int ii=0; ii<3; ii++) {
    currentFits = fits[ii];
    currentScale = &frScale[ii];
    channelFits = &fits[ii];
    currentCount = &fitsCount[ii];
    currentMode = &displayMode[ii];
    currentMosaic = &mosaicType[ii];

    currentScale->setClipScope(FrScale::LOCAL);
  
    if (!currentFits->hasDATAMIN() && currentScale->mmMode()==FrScale::DATAMIN)
      currentScale->setMMMode(FrScale::SAMPLE);
    if (!currentFits->hasIRAFMIN() && currentScale->mmMode()==FrScale::IRAFMIN)
      currentScale->setMMMode(FrScale::SAMPLE);

    updateClip();
    initHistogram();
    updateHistogram();
  }

  channel = 0;
  slice = 1;
  currentFits = fits[channel];
  currentScale = &frScale[channel];

  channelFits = &fits[channel];
  currentCount = &fitsCount[channel];
  currentMode = &displayMode[channel];
  currentMosaic = &mosaicType[channel];

  currentWCScdelt = &WCScdelt[channel];
  currentWCSrot = &WCSrot[channel];
  currentWCSorientation = &WCSorientation[channel];
  currentWCSbb = &WCSbb[channel];

  currentcontour = &contour[channel];

  resetValues();

  update(MATRIX);
}

void FrameRGB::getInfoSingle(const Vector& v, InternalSystem ref,
			     Matrix& (FitsImage::*matrixToData)(),
			     Matrix& (FitsImage::*matrixToImage)(),
			     SkyFrame sky, SkyFormat format, char* var)
{
  FrameBase::getInfoSingle(v,ref,matrixToData,matrixToImage,sky,format,var);

  // set RGB values
  char* array[3] = {"value,red","value,green","value,blue"};
  for (int ii=0; ii<3; ii++) {
    if (cfits[ii]) {
      Vector img = v * (cfits[ii]->*matrixToData)();
      int* params = cfits[ii]->getDataParams(currentScale->scanMode());
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];
	
      if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax)
	Tcl_SetVar2(interp,var,array[ii],(char*)cfits[ii]->getValue(img),0);
    }
    else
      Tcl_SetVar2(interp,var,array[ii],"",0);
  }
}

void FrameRGB::getInfoMosaic(const Vector& v, InternalSystem ref,
			     Matrix& (FitsImage::*matrixToData)(),
			     Matrix& (FitsImage::*matrixToImage)(),
			     SkyFrame sky, SkyFormat format, char* var)
{
  FrameBase::getInfoMosaic(v,ref,matrixToData,matrixToImage,sky,format,var);

  // set RGB values

  char* array[3] = {"value,red","value,green","value,blue"};
  for (int ii=0; ii<3; ii++) {
    FitsImage* ptr = fits[ii];
    while (ptr) {
      Vector img = v * (ptr->*matrixToData)();
      int* params = ptr->getDataParams(currentScale->scanMode());
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	Tcl_SetVar2(interp,var,array[ii],(char*)ptr->getValue(img),0);
	break;
      }
      ptr = ptr->next();
    }
  }
}

